from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import List, Optional
from uuid import UUID
from sqlalchemy import create_engine, select, update
from sqlalchemy.orm import Session
from credici.exception import InsufficientFundsError, UserDoesNotExistError
from credici.sync_.storage.storage import Storage
from credici.schema.postgres_db_schema import DB_Ledger
from credici.schema.postgres_db_schema import DB_UserCycle
from credici.schema.postgres_db_schema import DB_Users
from sqlalchemy.dialects.postgresql import insert


class PostgresStorage(Storage):
    """
    The PostgresStorage class for managing transactions.

    Attributes:
        url (str): The URL of the storage.
        auto_commit (bool): A flag to enable/disable auto-commit. If True, transactions are automatically committed.
            If False, users can manually manage transactions using the provided methods. The session starts when the class is instantiated.

    """

    url: str
    auto_commit: bool = True

    def __init__(self, url: str, auto_commit: Optional[bool] = True):
        self.url = url
        self.engine = create_engine(self.url)
        self.auto_commit = auto_commit
        self._session = None

        if self.auto_commit is False:
            self._session = Session(self.engine)

    def add_transaction(self, user_id: UUID, amount: int, expire_at: datetime) -> None:
        """
        Adds a transaction to the storage.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of the transaction.
        """
        if self.auto_commit is False:
            try:
                self._session.execute(
                    insert(DB_Ledger).values(
                        user_id=user_id,
                        amount=amount,
                        created_at=datetime.now(),
                        expire_at=expire_at,
                    )
                )
            except Exception:
                self._session.rollback()
                raise
        else:
            with Session(self.engine) as session:
                try:
                    session.execute(
                        insert(DB_Ledger).values(
                            user_id=user_id,
                            amount=amount,
                            created_at=datetime.now(),
                            expire_at=expire_at,
                        )
                    )
                    session.commit()
                except Exception:
                    session.rollback()
                    raise

    def get_active_ledger(self, user_id: UUID) -> DB_Ledger:
        with Session(self.engine) as session:
            try:
                stmt = (
                    select(DB_Ledger)
                    .where(DB_Ledger.user_id == user_id)
                    .where(DB_Ledger.expire_at > datetime.now())
                    .order_by(DB_Ledger.timestamp)
                )
                q = session.execute(stmt)
                if len(q) > 0:
                    return q.scalars().all()
                else:
                    return False
            except Exception:
                raise

    def get_all_user_transactions(self, user_id: UUID) -> List[DB_Ledger]:
        """
        Retrieves all the transactions for a specific user.
        """
        with Session(self.engine) as session:
            try:
                stmt = (
                    select(DB_Ledger)
                    .where(DB_Ledger.user_id == user_id)
                    .order_by(DB_Ledger.timestamp)
                )
                q = session.execute(stmt)
                return q.scalars().all()
            except Exception:
                raise

    def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> None:
        """
        Retrieves the transactions for a specific user within a given period.
        """
        with Session(self.engine) as session:
            try:
                stmt = (
                    select(DB_Ledger)
                    .where(DB_Ledger.user_id == user_id)
                    .where(DB_Ledger.timestamp >= from_)
                    .where(DB_Ledger.timestamp <= to_)
                    .order_by(DB_Ledger.timestamp)
                )
                q = session.execute(stmt)
                return q.scalars().all()
            except Exception:
                raise


    def start_cycle(
        self, amount: int, time_delta: int, user_id: Optional[UUID] = None
    ) -> None:
        """
        Starts a new cycle for a user whether it exists or not.

        Args:
            amount (int): The amount of the transaction.
            time_delta (int): The duration of the transaction in days.
            user_id (Optional[UUID], optional): The ID of the user. Defaults to None.

        """
        datetime_now = (
            datetime.now()
        )  # we store the current date and time in a variable so it's the same across the function
        if user_id is None:
            with Session(self.engine) as session:
                try:
                    new_user_id = self._add_new_user()
                    session.execute(
                        insert(DB_Ledger).values(
                            user_id=new_user_id,
                            amount=amount,
                            created_at=datetime_now,
                            expire_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    session.execute(
                        insert(DB_UserCycle).values(
                            user_id=new_user_id,
                            cycle_start_at=datetime_now,
                            cycle_end_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    session.commit()
                except Exception:
                    session.rollback()
                    raise
        else:
            if self.get_user_by_id(user_id) is None:
                raise UserDoesNotExistError(user_id)

            with Session(self.engine) as session:
                try:
                    session.execute(
                        insert(DB_Ledger).values(
                            user_id=user_id,
                            amount=amount,
                            timestamp=datetime_now,
                            created_at=datetime_now,
                            expire_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    session.execute(
                        insert(DB_UserCycle)
                        .values(
                            user_id=user_id,
                            cycle_start_at=datetime_now,
                            cycle_end_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                        .on_conflict_do_update(
                            index_elements=["user_id"],
                            set_={
                                "cycle_start_at": datetime_now,
                                "cycle_end_at": datetime_now.replace(
                                    hour=23, minute=59, second=59
                                )
                                + timedelta(days=time_delta),
                            },
                        ),
                    )
                    session.commit()
                except Exception:
                    session.rollback()
                    raise

    def _get_cycle(self, user_id: UUID) -> DB_UserCycle | None:
        with Session(self.engine) as session:
            try:
                stmt = select(DB_UserCycle).where(DB_UserCycle.user_id == user_id)
                result = session.execute(stmt)
                return result.scalar_one_or_none()
            except Exception:
                raise

    def _add_new_user(self):
        if self.auto_commit is False:
            try:
                stmt = insert(DB_Users).values().returning(DB_Users.id)
                new_user_id = self._session.execute(stmt).scalar()
                return new_user_id
            except Exception:
                self._session.rollback()
                raise
        else:
            with Session(self.engine) as session:
                try:
                    stmt = insert(DB_Users).values().returning(DB_Users.id)
                    new_user_id = session.execute(stmt).scalar()
                    session.commit()
                    return new_user_id
                except Exception:
                    session.rollback()
                    raise

    def get_user_by_id(self, user_id: UUID) -> DB_Users | None:
        """
        Retrieves a user by its ID."""
        with Session(self.engine) as session:
            try:
                stmt = select(DB_Users).where(DB_Users.id == user_id)
                result = session.execute(stmt)
                return result.scalar_one_or_none()
            except Exception:
                raise

    def commit_to_storage(self) -> None:
        """
        Commit the current session to the storage.
        """
        if self.auto_commit is False:
            try:
                self._session.commit()
            except Exception:
                self._session.rollback()
                raise
            finally:
                self._session.close()

    def rollback_from_storage(self) -> None:
        """
        Rollback the current session from the storage.
        """
        if self.auto_commit is False:
            try:
                self._session.rollback()
            except Exception:
                raise
            finally:
                self._session.close()



from pydantic import BaseModel
from typing import List
from datetime import datetime, timedelta
from uuid import UUID
from credici.exception import (
    AmountNegativeError,
    InsufficientFundsError,
    UserDoesNotExistError,
)
from credici.async_.async_storage.async_storage import AsyncStorage
from credici.sync_.credit_transaction import CreditTransaction
from pydantic import ConfigDict


class AsyncLedger(BaseModel):
    """The ledger class for managing credit transactions asynchronously."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    transactions: List[CreditTransaction] = []
    storage: AsyncStorage

    async def add(self, user_id: UUID, amount: int, expire_days: int) -> None:
        """Add credits for a user.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The amount of credits to add to the user's balance.
            expire_at (datetime, optional): The expiration date of the credits. Defaults to None.

        """
        if amount < 0:
            raise AmountNegativeError(amount)

        if await self.storage.get_user_by_id(user_id) is None:
            raise UserDoesNotExistError(user_id)
        
        expire_at = datetime.now().replace(hour=23, minute=59) + timedelta(
            days=expire_days
        )

        await self.storage.add_transaction(user_id, amount, expire_at)

    async def subtract(self, user_id: UUID, amount: int) -> None:
        """Subtract credits from a user.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The amount of credits to subtract from the user's balance.
        """
        if amount < 0:
            raise AmountNegativeError(amount)

        if await self.storage.get_user_by_id(user_id) is None:
            raise UserDoesNotExistError(user_id)
        
        active_balance = await self.get_active_balance(user_id)
        if active_balance < amount:
            raise InsufficientFundsError(user_id, active_balance, amount)
        
        expire_at = datetime.now()

        await self.storage.add_transaction(user_id, -amount, expire_at)

    async def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> List[CreditTransaction]:
        """Get all transactions for a user within a given period.

        Args:
            user_id (UUID): The unique identifier of the user.
            from_ (datetime): The start of the period.
            to_ (datetime): The end of the period.

        Returns:
            List[CreditTransaction]: A list of all transactions for the user within the given period.
        """
        if await self.storage.get_user_by_id(user_id) is None:
            raise UserDoesNotExistError(user_id)

        transactions = await self.storage.get_user_transactions_for_period(
            user_id, from_, to_
        )
        self.transactions = [
            CreditTransaction(
                transaction_id=transaction.transaction_id,
                user_id=transaction.user_id,
                amount=transaction.amount,
                timestamp=transaction.timestamp,
                created_at=transaction.created_at,
                expire_at=transaction.expire_at,
            )
            for transaction in transactions
        ]
        return self.transactions

    async def get_all_user_transactions(self, user_id: UUID) -> List[CreditTransaction]:
        """Get all transactions for a user.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            List[CreditTransaction]: A list of all transactions for the user.
        """
        if await self.storage.get_user_by_id(user_id) is None:
            raise UserDoesNotExistError(user_id)

        transactions = await self.storage.get_all_user_transactions(user_id)
        self.transactions = [
            CreditTransaction(
                transaction_id=transaction.transaction_id,
                user_id=transaction.user_id,
                amount=transaction.amount,
                timestamp=transaction.timestamp,
                created_at=transaction.created_at,
                expire_at=transaction.expire_at,
            )
            for transaction in transactions
        ]
        return self.transactions
    

    async def get_active_balance(self, user_id: UUID) -> int:
        """Get the active balance for a user.
        
        Args:
            user_id (UUID): The unique identifier of the user.
            
        Returns:
            int: The active balance for the user.
        """
        if await self.storage.get_user_by_id(user_id) is None:
            raise UserDoesNotExistError(user_id)
        try:
            active_ledger = await self.storage.get_active_ledger(user_id)
            active_balance = 0
            for transaction in active_ledger:
                active_balance += transaction.amount
            return active_balance
        except Exception:
            raise




from uuid import UUID
from datetime import datetime

"""Base schema for the storage layer."""


class StorageUser:
    id: UUID


class StorageUserCycle:
    id: UUID
    user_id: UUID
    cycle_start_at: datetime
    cycle_end_at: datetime


class StorageLedger:
    transaction_id: UUID
    user_id: UUID
    amount: int
    timestamp: datetime
    created_at: datetime
    expire_at: datetime



from datetime import datetime
from uuid import UUID
from credici.async_.async_storage.async_storage import AsyncStorage
from credici.async_.ledger import AsyncLedger
from credici.async_.user_cycle import AsyncUserCycle
from typing import List, Optional
from credici.sync_.credit_transaction import CreditTransaction


class AsyncCrediciClient:
    def __init__(self, storage: AsyncStorage):
        self.storage = storage
        self.user_balance = AsyncUserCycle(storage=self.storage)
        self.ledger = AsyncLedger(storage=self.storage)

    async def add(self, user_id: UUID, amount: int, expire_days: int = 30) -> None:
        """Add credits to a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of credits to add.
            expire_at (datetime): The time at which the credits will expire.
        """
        await self.ledger.add(user_id, amount, expire_days)

    async def subtract(self, user_id: UUID, amount: int) -> None:
        """Subtract credits from a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of credits to subtract.
        """
        await self.ledger.subtract(user_id, amount)

    async def get_balance(self, user_id: UUID) -> AsyncUserCycle:
        """Get the balance of a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            AsyncUserBalance: The user balance object of the given user.
        """
        return await self.ledger.get_active_balance(user_id)

    async def start_cycle(
        self, amount: int, time_delta: int = 30, user_id: Optional[UUID | None] = None
    ) -> None:
        """Start a new cycle for a user asynchronously.

        Args:
            amount (int): The amount of credits to add to the user's balance.
            time_delta (int): The duration of the cycle in days. Defaults to 30.
            user_id (Optional[UUID], optional): The ID of the user. If None, a new user will be created.
        """
        await self.user_balance.start_cycle(amount, time_delta, user_id)

    async def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> List[CreditTransaction]:
        """Get all transactions for a user within a given period asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
            from_ (datetime): The start of the period.
            to_ (datetime): The end of the period.
        """
        return await self.ledger.get_user_transactions_for_period(user_id, from_, to_)

    async def get_all_user_transactions(self, user_id: UUID) -> List[CreditTransaction]:
        """Get all transactions for a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
        """
        return await self.ledger.get_all_user_transactions(user_id)

    async def commit_to_storage(self) -> None:
        """Commit changes to the storage layer asynchronously if auto_commit=False."""
        await self.storage.commit_to_storage()

    async def rollback_from_storage(self) -> None:
        """Rollback changes to the storage layer asynchronously if auto_commit=False."""
        await self.storage.rollback_from_storage()



from contextlib import asynccontextmanager, contextmanager
from datetime import datetime, timedelta
from uuid import UUID
from credici.schema.postgres_db_schema import DB_Ledger, DB_Users, DB_UserCycle
from sqlalchemy import select, update
from credici.exception import InsufficientFundsError, UserDoesNotExistError
from credici.async_.async_storage.async_storage import AsyncStorage
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from typing import List, Optional
from sqlalchemy.dialects.postgresql import insert


class AsyncPostgresStorage(AsyncStorage):
    """
    The AsyncPostgresStorage class for managing transactions asynchronously.

    Attributes:
        url (str): The URL of the storage.
        auto_commit (bool): A flag to enable/disable auto-commit. If True, transactions are automatically committed.
            If False, users can manually manage transactions using the provided methods. The session starts when the class is instantiated.
    """

    url: str
    auto_commit: bool = True

    def __init__(self, url: str, auto_commit: Optional[bool] = True):
        self.url = url
        self.engine = create_async_engine(url, echo=True)
        self.auto_commit = auto_commit
        self._asession = None

        if self.auto_commit is False:
            self._asession = AsyncSession(self.engine)

    async def add_transaction(
        self, user_id: UUID, amount: int, expire_at: datetime = None
    ) -> None:
        """Adds a transaction to the storage asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of the transaction.
        """
        if self.auto_commit is False:
            try:
                self._asession.add(
                    insert(DB_Ledger).values(
                        user_id=user_id,
                        amount=amount,
                        created_at=datetime.now(),
                        expire_at=expire_at,
                    )
                )
            except Exception:
                self._asession.rollback()
                raise
        else:
            async with AsyncSession(self.engine) as asession:
                try:
                    asession.add(
                        insert(DB_Ledger).values(
                            user_id=user_id,
                            amount=amount,
                            created_at=datetime.now(),
                            expire_at=expire_at,
                        )
                    )
                    await asession.commit()
                except Exception:
                    await asession.rollback()
                    raise

    async def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> List[DB_Ledger]:
        """Get all transactions for a user within a given period asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.
            from_ (datetime): The start of the period.
            to_ (datetime): The end of the period.

        Returns:
            List[DB_Ledger]: A list of transactions for the user within the given period.
        """
        async with AsyncSession(self.engine) as asession:
            result = await asession.execute(
                select(DB_Ledger)
                .where(DB_Ledger.user_id == user_id)
                .where(DB_Ledger.created_at >= from_)
                .where(DB_Ledger.created_at <= to_)
            )
            return result.scalars().all()

    async def get_all_user_transactions(self, user_id: UUID) -> List[DB_Ledger]:
        """Get all transactions for a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            List[DB_Ledger]: A list of all transactions for the user.
        """
        async with AsyncSession(self.engine) as asession:
            try:
                result = await asession.execute(
                    select(DB_Ledger).where(DB_Ledger.user_id == user_id)
                )
                return result.scalars().all()
            except Exception:
                raise

    async def get_active_balance(self, user_id: UUID) -> int | None:
        """Get the balance of a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            int: The balance of the user.
        """
        try:
            active_ledger = await self.get_active_ledger(user_id)
            if active_ledger is not None:
                for t in active_ledger:
                    active_balance = +t.amount
                return active_balance
            else:
                return None
        except Exception:
            raise

    async def start_cycle(
        self, amount: int, time_delta: int = 30, user_id: Optional[UUID | None] = None
    ) -> None:
        """Start a new cycle for a user asynchronously.

        Args:

            amount (int): The amount of credits to add to the user's balance.
            time_delta (int): The duration of the cycle in days. Defaults to 30.
            user_id (Optional[UUID], optional): The ID of the user. If None a new user will be created.
        """
        datetime_now = (
            datetime.now()
        )  # we store the current date and time in a variable so it's the same across the function

        if user_id is None:
            async with AsyncSession(self.engine) as asession:
                try:
                    new_user_id = self._add_new_user()
                    asession.add(
                        DB_Ledger(
                            user_id=new_user_id,
                            amount=amount,
                            created_at=datetime_now,
                            expire_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    await asession.execute(
                        insert(DB_UserCycle).values(
                            user_id=new_user_id,
                            cycle_start_at=datetime_now,
                            cycle_end_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    await asession.commit()
                except Exception:
                    raise
        else:
            if self.get_user_by_id(user_id) is None:
                raise UserDoesNotExistError(user_id)
            async with AsyncSession(self.engine) as asession:
                try:
                    asession.add(
                        DB_Ledger(
                            user_id=user_id,
                            amount=amount,
                            timestamp=datetime_now,
                            created_at=datetime_now,
                            expire_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                    )
                    await asession.execute(
                        insert(DB_UserCycle)
                        .values(
                            user_id=user_id,
                            cycle_start_at=datetime_now,
                            cycle_end_at=datetime_now.replace(
                                hour=23, minute=59, second=59
                            )
                            + timedelta(days=time_delta),
                        )
                        .on_conflict_do_update(
                            index_elements=["user_id"],
                            set_={
                                "balance": amount,
                                "cycle_start_at": datetime_now,
                                "cycle_end_at": datetime_now.replace(
                                    hour=23, minute=59, second=59
                                )
                                + timedelta(days=time_delta),
                            },
                        ),
                    )
                    await asession.commit()
                except Exception:
                    await asession.rollback()
                    raise

    async def _add_new_user(self) -> UUID:
        """Add a new user to the storage asynchronously.

        Returns:
            UUID: The unique identifier of the new user.
        """
        if self.auto_commit is False:
            try:
                new_user_id = await self._asession.execute(
                    insert(DB_Users).values().returning(DB_Users.id)
                )
                return new_user_id.scalar()
            except Exception:
                await self._asession.rollback()
                raise
        else:
            async with AsyncSession(self.engine) as asession:
                try:
                    new_user_id = await asession.execute(
                        insert(DB_Users).values().returning(DB_Users.id)
                    )
                    await asession.commit()
                    return new_user_id.scalar()
                except Exception:
                    await asession.rollback()
                    raise

    async def get_user_by_id(self, user_id: UUID) -> DB_Users:
        """Get a user by ID asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            DB_Users: The user object.
        """
        async with AsyncSession(self.engine) as asession:
            try:
                result = await asession.execute(
                    select(DB_Users).where(DB_Users.id == user_id)
                )
                return result.scalar_one_or_none()
            except Exception:
                raise

    async def commit_to_storage(self) -> None:
        """Commit changes to the storage layer if auto_commit=False."""
        if self.auto_commit is False:
            try:
                await self._asession.commit()
            except Exception:
                await self._asession.rollback()
                raise

    async def rollback_from_storage(self) -> None:
        """Rollback changes to the storage layer if auto_commit=False."""
        if self.auto_commit is False:
            try:
                await self._asession.rollback()
            except Exception:
                raise
            finally:
                await self._asession.close()

    async def get_active_ledger(self, user_id: UUID) -> DB_Ledger:
        """Get the current active ledger for a user asynchronously.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            DB_Ledger: The current active ledger for the user.
        """
        async with AsyncSession(self.engine) as asession:
            try:
                result = await asession.execute(
                    select(DB_Ledger)
                    .where(DB_Ledger.user_id == user_id)
                    .where(DB_Ledger.expire_at > datetime.now())
                )
                return result.scalars().all()
            except Exception:
                raise



from turtle import back
from sqlalchemy.orm import DeclarativeBase, relationship, MappedAsDataclass
from sqlalchemy import ForeignKey, text
from sqlalchemy.orm import Mapped, mapped_column
from datetime import datetime
from uuid import UUID
from typing import Any, Dict, List


class Base(MappedAsDataclass, DeclarativeBase):
    """Base class for SQLAlchemy models."""

    pass


class DB_Users(Base):
    __tablename__ = "users"

    id: Mapped[UUID] = mapped_column(
        primary_key=True, server_default=text("uuid_generate_v4()")
    )


class DB_UserCycle(Base):
    __tablename__ = "user_cycle"

    id: Mapped[UUID] = mapped_column(
        primary_key=True, server_default=text("uuid_generate_v4()")
    )
    user_id: Mapped[UUID] = mapped_column(ForeignKey("users.id"))
    cycle_start_at: Mapped[datetime]
    cycle_end_at: Mapped[datetime]


class DB_Ledger(Base):
    __tablename__ = "ledger"

    transaction_id: Mapped[UUID] = mapped_column(
        primary_key=True, server_default=text("uuid_generate_v4()")
    )
    user_id: Mapped[UUID] = mapped_column(ForeignKey("users.id"))
    amount: Mapped[int]
    timestamp: Mapped[datetime] = mapped_column(server_default=text("now()"))
    created_at: Mapped[datetime]
    expire_at: Mapped[datetime]



from abc import abstractmethod
from uuid import UUID
from typing import List, Optional
from uuid import UUID
from datetime import datetime
from credici.schema.storage_schema import StorageLedger, StorageUser, StorageUserCycle


class AsyncStorage:
    @abstractmethod
    async def add_transaction(
        self, user_id: UUID, amount: int, expire_at: datetime 
    ) -> None:
        pass

    @abstractmethod
    async def get_active_ledger(self, user_id: UUID) -> List[StorageLedger]:
        pass

    @abstractmethod
    async def get_all_user_transactions(self, user_id: UUID) -> List[StorageLedger]:
        pass

    @abstractmethod
    async def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> List[StorageLedger]:
        pass

    @abstractmethod
    async def start_cycle(
        self, amount: int, time_delta: int, user_id: Optional[UUID | None] = None
    ) -> None:
        pass

    @abstractmethod
    async def get_user_by_id(self, user_id: UUID) -> StorageUser:
        pass

    @abstractmethod
    async def commit_to_storage(self) -> None:
        pass

    @abstractmethod
    async def rollback_from_storage(self) -> None:
        pass



from datetime import datetime
from uuid import UUID
from credici.async_.ledger import AsyncLedger
from credici.async_.async_storage.async_storage import AsyncStorage
from typing import Optional
from pydantic import BaseModel, ConfigDict, Field


class AsyncUserCycle(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: UUID = None
    user_id: UUID = None
    cycle_start_at: datetime = None
    cycle_end_at: datetime = None
    ledger: AsyncLedger = None
    storage: AsyncStorage = Field(exclude=True)

    async def start_cycle(
        self, amount: int, time_delta: int = 30, user_id: Optional[UUID | None] = None
    ):
        """Start a new cycle for a user asynchronously.

        Args:
            amount (int): The amount of credits to add to the user's balance.
            time_delta (int): The duration of the cycle in days. Defaults to 30.
            user_id (Optional[UUID], optional): The ID of the user. If None a new user will be created.
        """
        await self.storage.start_cycle(amount, time_delta, user_id)



from datetime import datetime
from uuid import UUID
from credici.sync_.credit_transaction import CreditTransaction
from credici.sync_.ledger import Ledger
from credici.sync_.user_cycle import UserCycle
from credici.sync_.storage.storage import Storage
from typing import List, Optional


class CrediciClient:
    def __init__(self, storage: Storage):
        self.storage = storage
        self.user_balance = UserCycle(storage=self.storage)
        self.ledger = Ledger(storage=self.storage)

    def add(self, user_id: UUID, amount: int, expire_days: int = 30) -> None:
        """Add credits to a user.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of credits to add.
            expire_days (int): The number of days until the credits expire.
        """
        return self.ledger.add(user_id, amount, expire_days)

    def subtract(self, user_id: UUID, amount: int) -> None:
        """Subtract credits from a user.

        Args:
            user_id (UUID): The unique identifier of the user.
            amount (int): The absolute amount of credits to subtract.
        """
        return self.ledger.subtract(user_id, amount)

    def get_balance(self, user_id: UUID) -> int:
        """Get the balance of a user.

        Args:
            user_id (UUID): The unique identifier of the user.

        Returns:
            UserBalance: The user balance object of the given user.
        """
        return self.ledger.get_active_balance(user_id)

    def start_cycle(
        self, amount: int, time_delta: int = 30, user_id: Optional[UUID | None] = None
    ) -> None:
        """Start a new cycle for a user.

        Args:

            amount (int): The amount of credits to add to the user's balance.
            time_delta (int): The duration of the cycle in days. Defaults to 30.
            user_id (Optional[UUID], optional): The ID of the user. If None a new user will be created.
        """
        return self.user_balance.start_cycle(amount, time_delta, user_id)

    def get_user_transactions_for_period(
        self, user_id: UUID, from_: datetime, to_: datetime
    ) -> List[CreditTransaction]:
        """Get all transactions for a user within a given period.

        Args:
            user_id (UUID): The unique identifier of the user.
            from_ (datetime): The start of the period.
            to_ (datetime): The end of the period.
        """
        return self.ledger.get_user_transactions_for_period(user_id, from_, to_)

    def get_all_user_transactions(self, user_id: UUID) -> List[CreditTransaction]:
        """Get all transactions for a user.

        Args:
            user_id (UUID): The unique identifier of the user.
        """
        return self.ledger.get_all_user_transactions(user_id)

    def commit_to_storage(self) -> None:
        """Commit changes to the storage layer if auto_commit=False."""
        return self.storage.commit_to_storage()

    def rollback_from_storage(self) -> None:
        """Rollback changes to the storage layer if auto_commit=False."""
        return self.storage.rollback_from_storage()



from datetime import datetime
from uuid import UUID


class InsufficientFundsError(Exception):
    amount: int
    user_id: UUID

    def __init__(self, user_id: UUID, amount: int):
        self.amount = amount
        self.user_id = user_id

        super().__init__(
            f"User {user_id} does not have enough funds to subtract {amount}."
        )


class AmountNegativeError(Exception):
    amount: int

    def __init__(self, amount: int):
        self.amount = amount
        super().__init__(f"Amount {amount} is negative. Provide a positive amount.")


class UserDoesNotExistError(Exception):
    user_id: UUID

    def __init__(self, user_id: UUID):
        self.user_id = user_id
        super().__init__(f"User {user_id} does not exist.")


class CreditExpiredError(Exception):
    user_id: UUID
    date: datetime

    def __init__(self, user_id: UUID, date: datetime):
        self.user_id = user_id
        self.date = date
        super().__init__(f"User {user_id} credit has expired on {date}")



from datetime import datetime
from uuid import UUID
from pydantic import BaseModel


class CreditTransaction(BaseModel):
    """
    Represents a ledger entry for a transaction.

    Attributes:
        transaction_id (UUID or None): The unique identifier of the transaction.
        user_id (UUID): The unique identifier of the user associated with the transaction.
        amount (int): The absolute amount of the transaction.
        timestamp (str or None): The timestamp of the transaction.
    """

    transaction_id: UUID 
    user_id: UUID
    amount: int
    timestamp: datetime
    created_at: datetime
    expire_at: datetime
